#!python
# coding=utf-8
import simplejson as json
from datetime import datetime

import pytz
import sqlalchemy as sql
from sqlalchemy.dialects.postgresql import JSONB

from dbsink import L


def payload_parse(payload):
    # Make sure we have valid JSON and remove any
    # Infinity and NaN values in the process
    try:
        return json.loads(json.dumps(payload, ignore_nan=True))
    except BaseException as e:
        raise ValueError(f'Could not parse message as valid JSON - {repr(e)}')


class BaseMap:
    def __init__(self, topic, **kwargs):
        self.topic = topic
        self.table = kwargs.get('table', topic).replace('.', '-')
        self.filters = kwargs.get('filters', {})

    @property
    def upsert_constraint_name(self):
        return f'{self.table}_unique_constraint'.replace('-', '_').lower()

    @property
    def unique_index_name(self):
        return f'{self.table}_unique_idx'.replace('-', '_').lower()

    @property
    def sequence_name(self):
        return f'{self.table}_id_seq'.replace('-', '_').lower()

    def _check_key(self, key):
        """ Check if the key is valid
            return True if valid
            raise a KeyError if invalid
        """
        raise NotImplementedError

    def _check_value(self, value):
        """ Check if the key is valid
            return True if valid
            raise a ValueError if invalid
        """
        raise NotImplementedError

    def check(self, key, value):
        """ Check if the key or value is valid """
        return self._check_key(key) is True and self._check_value(value) is True

    @property
    def schema(self):
        """ Return the SQLAlchemy table schema as well as the
            unique index check to upsert data
        """
        raise NotImplementedError

    def match_columns(self, inserts):
        """ Throws away insert data that does not match a defined
            column name.
        """
        # Get column names
        column_names = [ s.name for s in self.schema if isinstance(s, sql.Column) ]
        # Throw away keys that are not column names
        matched_inserts = { k: v for k, v in inserts.items() if k in column_names }

        if inserts.keys() != matched_inserts.keys():
            unmatched = [ x for x in inserts.keys() if x not in matched_inserts ]
            L.info(f'Threw away data with no columns: {unmatched}')

        return matched_inserts

    def message_to_values(self, key, value):
        raise NotImplementedError


class JsonMap(BaseMap):

    @property
    def upsert_constraint_name(self):
        return None

    def _check_key(self, key):
        return True

    def _check_value(self, value):
        _ = payload_parse(value)
        return True

    @property
    def schema(self):
        return [
            sql.Column('id',       sql.Integer, sql.Sequence(self.sequence_name), primary_key=True),
            sql.Column('sinked',   sql.DateTime(timezone=True), index=True),
            sql.Column('key',      sql.String, default='', index=True),
            sql.Column('payload',  JSONB)
        ]

    def message_to_values(self, key, value):

        # Raises if invalid
        value = payload_parse(value)

        self.check(key, value)

        values = {
            'sinked':  datetime.utcnow().replace(tzinfo=pytz.utc).isoformat(),
            'key':     key,
            'payload': value,
        }

        return key, values


class StringMap(BaseMap):

    @property
    def upsert_constraint_name(self):
        return None

    def _check_key(self, key):
        return True

    def _check_value(self, value):
        return True

    @property
    def schema(self):
        return [
            sql.Column('id',       sql.Integer, sql.Sequence(self.sequence_name), primary_key=True),
            sql.Column('sinked',   sql.DateTime(timezone=True), index=True),
            sql.Column('key',      sql.String, default='', index=True),
            sql.Column('payload',  sql.String)
        ]

    def message_to_values(self, key, value):

        # Raises if invalid
        self.check(key, value)

        values = {
            'sinked':  datetime.utcnow().replace(tzinfo=pytz.utc).isoformat(),
            'key':     key,
            'payload': json.dumps(value),
        }

        return key, values



#!python
# coding=utf-8
import uuid
import simplejson as json

import msgpack

from easyavro import EasyAvroConsumer, EasyConsumer

from dbsink import L


class MessageFiltered(Exception):
    pass


def get_kafka_consumer(brokers, topic, offset, packing, consumer=None, registry=None):

    # Generate a random consumer if one was not provided.
    # This guarentees a unique consumer ID for each run
    if not consumer:
        consumer = f'dbsink-{topic}-{uuid.uuid4().hex[0:20]}'
        L.info(f'Setting consumer to {consumer}')

    consumer_kwargs = {
        'kafka_brokers': brokers,
        'consumer_group': consumer,
        'kafka_topic': topic,
        'offset': offset
    }

    # Setup the kafka consuimer
    if packing == 'avro':
        unpacking_func = None
        packing_func = None
        consumer_class = EasyAvroConsumer
        if not registry:
            raise ValueError('Avro packing requestd but no schema registry url was found!')
        consumer_kwargs.update({
            'schema_registry_url': registry,
        })
    elif packing == 'msgpack':
        unpacking_func = lambda x: msgpack.loads(x, use_list=False, raw=False)  # noqa
        packing_func = lambda x: msgpack.packb(x, use_bin_type=True)  # noqa
        consumer_class = EasyConsumer
    elif packing == 'json':
        unpacking_func = json.loads
        packing_func = lambda x: json.dumps(x, ignore_nan=True)  # noqa
        consumer_class = EasyConsumer

    return consumer_class, consumer_kwargs, unpacking_func, packing_func


def listen_unpack(brokers, topic, offset, packing, mapping, consumer=None, registry=None, on_receive=None, loop=False):

    consume_cls, consume_kw, unpack, _ = get_kafka_consumer(
        brokers=brokers.split(','),
        topic=topic,
        offset=offset,
        packing=packing,
        consumer=consumer,
        registry=registry
    )

    if on_receive is None:
        def on_receive(k, v):
            L.info("Recieved message:\nKey: {}\nValue: {}".format(k, v))

    def unpack_receive(k, v):
        if v is not None and unpack:
            try:
                v = unpack(v)
            except BaseException:
                L.error(f'Error unpacking message using {packing}: {v}')
                return

        try:
            nk, nv = mapping.message_to_values(k, v)
        except MessageFiltered as e:
            L.debug(e)
            return
        except BaseException as e:
            L.error(f'Skipping {v} - {repr(e)}')
            return

        on_receive(nk, nv)

    c = consume_cls(**consume_kw)
    c.consume(
        on_recieve=unpack_receive,
        initial_wait=1,
        timeout=10,
        cleanup_every=1000,
        loop=loop
    )



#!python
# coding=utf-8
import logging


log_format = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
stream = logging.StreamHandler()
stream.setFormatter(log_format)

ea = logging.getLogger('easyavro')
ea.setLevel(logging.INFO)
ea.addHandler(stream)

L = logging.getLogger()
L.setLevel(logging.INFO)
L.handlers = [stream]

__version__ = "2.6.0"



#!python
# coding=utf-8
import logging
import pkg_resources
import simplejson as json
from datetime import datetime

import pytz
import click

import sqlalchemy as sql
from sqlalchemy.dialects.postgresql import insert

from dbsink import L, ea, log_format, utils


def get_mappings():
    return {
        e.name: e.resolve() for e in pkg_resources.iter_entry_points('dbsink.maps')
    }


@click.command()
@click.option('--brokers',  type=str, required=True, default='localhost:4001', help="Kafka broker string (comman separated).")
@click.option('--topic',    type=str, required=True, default='axds-netcdf-replayer-data', help="Kafka topic to send the data to. '-value' is auto appended if using avro packing.")
@click.option('--table',    type=str, required=False, default='', help="Name of the table to sink into. Defaults to the topic name.")
@click.option('--lookup',   type=str, required=False, default='JsonMap', help="Lookup name to use to find the correct table format (default: the topic name).")
@click.option('--db',       type=str, required=True, default='postgresql+psycopg2://sink:sink@localhost:30300/sink', help="SQLAlchemy compatible postgres connection string.")
@click.option('--schema',   type=str, required=True, default='public', help="Database schema to use (default: public).")
@click.option('--consumer', type=str, default='', help="Consumer group to listen with (default: random).")
@click.option('--offset',   type=str, required=True, default='largest', help="Kafka offset to start with (default: largest).")
@click.option('--packing',  type=click.Choice(['json', 'avro', 'msgpack']), default='json', help="The data unpacking algorithm to use (default: json).")
@click.option('--registry', type=str, default='http://localhost:4002', help="URL to a Schema Registry if avro packing is requested")
@click.option('--drop/--no-drop', default=False, help="Drop the table first")
@click.option('--truncate/--no-truncate', default=False, help="Truncate the table first")
@click.option('--logfile',  type=str, default='', help="File to log messages to (default: stdout).")
@click.option('--listen/--no-listen', default=True, help="Whether to listen for messages.")
@click.option('--do-inserts/--no-do-inserts', default=True, help="Whether to insert data into a database.")
@click.option('--datafile', type=str, default='', help="File to pull messages from instead of listening for messages.")
@click.option('-v', '--verbose', count=True, help="Control the output verbosity, use up to 3 times (-vvv)")
# Filters
@click.option('--start_date', type=click.DateTime(), required=False, default=None, help="Start date filter passed to each mapping class (UTC)")
@click.option('--end_date',   type=click.DateTime(), required=False, default=None, help="End date filter passed to each mapping class (UTC)")
def setup(brokers, topic, table, lookup, db, schema, consumer, offset, packing, registry, drop, truncate, logfile, listen, do_inserts, datafile, verbose, start_date, end_date):

    if logfile:
        handler = logging.FileHandler(logfile)
        handler.setFormatter(log_format)
        ea.addHandler(handler)
        L.addHandler(handler)

    if verbose == 0:
        ea.setLevel(logging.INFO)
        L.setLevel(logging.INFO)
    elif verbose >= 1:
        ea.setLevel(logging.DEBUG)
        L.setLevel(logging.DEBUG)

    # If no specific table was specified, use the topic name
    if not table:
        table = topic

    # Get consumer and unpack/pack information based on packing
    consume_cls, consume_kw, unpack, pack = utils.get_kafka_consumer(
        brokers=brokers.split(','),
        topic=topic,
        offset=offset,
        packing=packing,
        consumer=consumer,
        registry=registry
    )

    filters = {}
    if isinstance(start_date, datetime):
        filters['start_date'] = start_date.replace(tzinfo=pytz.utc)
    if isinstance(end_date, datetime):
        filters['end_date'] = end_date.replace(tzinfo=pytz.utc)

    # Get the mapping object from the lookup parameter
    mappings = get_mappings()
    mapping = mappings[lookup](topic, table=table, filters=filters)
    L.debug(f'Using mapping: {lookup}, topic: {topic}, table: {mapping.table}, filters: {len(filters)}')

    if do_inserts is True:
        """ Database connection and setup
        """
        engine = sql.create_engine(
            db,
            pool_size=5,
            max_overflow=100,
            pool_recycle=3600,
            pool_pre_ping=True,
            client_encoding='utf8',
            use_native_hstore=True,
            echo=verbose >= 2
        )
        # Create schema
        engine.execute(f"CREATE SCHEMA if not exists {schema}")

        # Add HSTORE extension
        engine.execute("CREATE EXTENSION if not exists hstore cascade")

        if drop is True:
            L.info(f'Dropping table {mapping.table}')
            engine.execute(sql.text(f'DROP TABLE IF EXISTS \"{mapping.table}\"'))

        # If we didn't drop the table, we should now truncate it.
        # There is no need to truncate if we just dropped the table.
        if drop is False and truncate is True:
            L.info(f'Truncating table {mapping.table}')
            try:
                engine.execute(sql.text(f'TRUNCATE TABLE \"{mapping.table}\" RESTART IDENTITY'))
            except BaseException as e:
                L.error(f'Could not truncate table: {e}')

        # Reflect to see if this table already exists. Create or update it.
        meta = sql.MetaData(engine, schema=schema)
        meta.reflect()
        if f'{schema}.{mapping.table}' not in meta.tables:
            sqltable = sql.Table(mapping.table, meta, *mapping.schema)
        else:
            sqltable = sql.Table(
                mapping.table,
                meta,
                *mapping.schema,
                autoload=True,
                keep_existing=False,
                extend_existing=True
            )
        meta.create_all(tables=[sqltable])

    def on_recieve(k, v):
        if v is not None and unpack:
            try:
                v = unpack(v)
            except BaseException:
                L.error(f'Error unpacking message using {packing}: {v}')
                return

        # Custom conversion function for the table
        try:
            newkey, newvalues = mapping.message_to_values(k, v)
        except utils.MessageFiltered as e:
            L.debug(e)
            return
        except BaseException as e:
            L.error(f'Skipping {v}, message could not be converted to a row - {repr(e)}')
            return

        if do_inserts:
            # I wonder if we can just do set_=v? Other seem to extract the
            # exact columns to update but this method is currently working...
            # https://gist.github.com/bhtucker/c40578a2fb3ca50b324e42ef9dce58e1
            insert_cmd = insert(sqltable).values(newvalues)
            if mapping.upsert_constraint_name is not None:
                upsert_cmd = insert_cmd.on_conflict_do_update(
                    constraint=mapping.upsert_constraint_name,
                    set_=newvalues
                )
                res = engine.execute(upsert_cmd)
                mode = 'inserted/updated'
            else:
                res = engine.execute(insert_cmd)
                mode = 'inserted'
            res.close()
            L.debug(f'{mode} row {res.inserted_primary_key}')

    if datafile:
        with open(datafile) as f:
            messages = json.load(f)
            for m in messages:
                on_recieve(None, pack(m))
    elif listen is True:
        c = consume_cls(**consume_kw)
        c.consume(
            on_recieve=on_recieve,
            initial_wait=1,
            timeout=10,
            cleanup_every=100,
            loop=True
        )


def run():
    setup(auto_envvar_prefix='DBSINK')


if __name__ == '__main__':
    run()



#!python
# coding=utf-8
import re
import ast
import collections
import simplejson as json
from datetime import datetime

import pytz
import sqlalchemy as sql
from shapely.ops import unary_union
from geoalchemy2.types import Geometry
from geoalchemy2.shape import from_shape
from shapely.geometry import shape, Point, box
from dateutil.parser import parse as dtparse
from sqlalchemy.dialects.postgresql import HSTORE, JSONB, DOUBLE_PRECISION

from dbsink.maps import BaseMap, payload_parse
from dbsink.utils import MessageFiltered
from dbsink import L  # noqa

xx = re.compile(r'[\x00-\x1f\\"]')
ux = re.compile(r'[\\u[0-9A-Fa-f]]')


WGS84_BBOX_180 = box(-180, -90, 180, 90)
WGS84_BBOX_360 = box(0, -90, 360, 90)


def flatten(d, parent_key='', sep='_'):
    # Adapted from:
    # https://stackoverflow.com/questions/6027558/flatten-nested-dictionaries-compressing-keys
    # to support nested lists and decoding of JSON objects that might be strings.
    items = []
    if isinstance(d, collections.MutableMapping):
        for k, v in d.items():
            new_key = f'{parent_key}{sep}{k}' if parent_key else k
            items.extend(
                flatten(v, new_key, sep=sep
            ).items())
    elif isinstance(d, list):
        # add the list object and its members as indexed items
        items.append((parent_key, d))
        for i, litem in enumerate(d):
            new_key = f'{parent_key}{sep}{i}' if parent_key else i
            items.extend(
                flatten(litem, new_key, sep=sep
            ).items())
    elif isinstance(d, str):
        # Try to expand it
        try:
            decoded = expand_json_objects(d)
            items.extend(
                flatten(decoded, parent_key, sep=sep
            ).items())
        except ValueError:
            # Not decodable, just set the objects value
            items.append((parent_key, d))
    else:
        items.append((parent_key, d))

    return dict(items)


def expand_json_objects(str_value):
    decoders = [
        json.loads,
        lambda x: json.loads(json.dumps(ast.literal_eval(x)))
    ]

    for d in decoders:
        try:
            # If this decoder object was successful, return it
            return d(str_value)
        except BaseException:
            continue

    # Return the original value if nothing was decoded
    raise ValueError('Could not decode any JSON object')


def get_point_location_quality(loc_geom, inprecise_location=False, disallow_lon=None, disallow_lat=None):
    """ QARTOD Flags:
        1 - Good
        2 - Not evaluated
        3 - Questionable/suspect
        4 - Bad
        9 - Missing Data
    """

    # Avoid locations that are both small decimal numbers.
    if -1 < loc_geom.x < 1 and -1 < loc_geom.y < 1:
        return 4

    # Avoid "null island" locations
    if loc_geom.x == 0 or loc_geom.y == 0:
        return 4

    # if we need to ignore certain X values
    if isinstance(disallow_lon, list) and loc_geom.x in disallow_lon:
        return 4

    # if we need to ignore certain Y values
    if isinstance(disallow_lat, list) and loc_geom.y in disallow_lat:
        return 4

    # Make sure we have resonable coordinates
    if not any([
        loc_geom.within(WGS84_BBOX_180),
        loc_geom.within(WGS84_BBOX_360)
    ]):
        return 4

    # If using an inprecise location (ie. Iridium)
    if inprecise_location is True:
        return 3

    return 1


def apply_start_end_filter(message_time, starting, ending):
    if isinstance(starting, datetime) and message_time < starting:
        raise MessageFiltered(f'Filtering out message from {message_time} since it is before {starting}')
    elif isinstance(ending, datetime) and message_time > ending:
        raise MessageFiltered(f'Filtering out message from {message_time} since it is after {ending}')


def make_valid_string(obj):
    if isinstance(obj, str):
        try:
            return ux.sub(
                '',
                xx.sub(
                    '',
                    obj
                )
            ).replace(
                '\x80',
                ''
            ).replace(
                '\x00',
                ''
            )
        except BaseException:
            return obj
    else:
        return str(obj)


class GenericFieldStatistic(BaseMap):

    @property
    def schema(self):
        return [
            sql.Column('id',       sql.Integer, sql.Sequence(self.sequence_name), primary_key=True),
            sql.Column('source',   sql.String, index=True, nullable=False),
            sql.Column('period',   sql.String, default=''),
            sql.Column('starting', sql.DateTime(timezone=True), index=True, nullable=False),
            sql.Column('ending',   sql.DateTime(timezone=True), index=True, nullable=False),
            sql.Column('values',   JSONB),
            sql.Index(
                self.unique_index_name,
                'source',
                'period',
                'starting',
                'ending',
                unique=True,
            ),
            sql.UniqueConstraint(
                'source',
                'period',
                'starting',
                'ending',
                name=self.upsert_constraint_name
            )
        ]

    def message_to_values(self, key, value):
        value = payload_parse(value)

        # Throw away non-column data
        value = self.match_columns(value)

        value['starting'] = dtparse(value['starting']).replace(tzinfo=pytz.utc)
        value['ending'] = dtparse(value['ending']).replace(tzinfo=pytz.utc)

        # Filter to make sure the time is represented in the filter window
        # First filter the `starting` between min and `end_date` filter
        # Then filter the `ending` between `start_date` filter and max
        # After both filters we are left with the defined window.
        apply_start_end_filter(
            value['starting'],
            datetime.min.replace(tzinfo=pytz.utc),
            self.filters.get('end_date')
        )
        apply_start_end_filter(
            value['ending'],
            self.filters.get('start_date'),
            datetime.max.replace(tzinfo=pytz.utc),
        )

        # Remove None to use the defaults defined in the table definition
        return key, { k: v for k, v in value.items() }


class GenericGeography(BaseMap):

    @property
    def schema(self):
        return [
            sql.Column('id',       sql.Integer, sql.Sequence(self.sequence_name), primary_key=True),
            sql.Column('uid',      sql.String, default='', index=True),
            sql.Column('gid',      sql.String, default='', index=True),
            sql.Column('time',     sql.DateTime(timezone=True), index=True),
            sql.Column('reftime',  sql.DateTime(timezone=True), index=True),
            sql.Column('values',   HSTORE, default={}),
            sql.Column('payload',  JSONB, default={}),
            sql.Column('geom',     Geometry(srid=4326)),
            sql.Index(
                self.unique_index_name,
                'uid',
                'gid',
                'time',
                unique=True,
            ),
            sql.UniqueConstraint(
                'uid',
                'gid',
                'time',
                name=self.upsert_constraint_name
            )
        ]

    def message_to_values(self, key, value):
        payload = payload_parse(value)

        tops = ['id', 'uid', 'gid', 'time', 'reftime', 'values', 'payload', 'geom', 'geojson']
        top_level = value.copy()

        top_time = dtparse(top_level['time']).replace(tzinfo=pytz.utc)
        apply_start_end_filter(
            top_time,
            self.filters.get('start_date'),
            self.filters.get('end_date')
        )

        # Save GeoJSON
        if isinstance(top_level['geojson'], str):
            geojson = json.loads(top_level['geojson'])
        else:
            geojson = top_level['geojson']

        features = []
        if 'features' in geojson:
            # This is a FeatureCollection
            features = geojson['features']
        elif 'coordinates' in geojson:
            # This is a geometry object, make a feature
            features = [{ "type": "Feature", "properties": {}, "geometry": geojson }]
        elif 'geometry' in geojson:
            # This is a Feature, cool.
            features = [geojson]

        del top_level['geojson']
        # Merge any geometries into one
        top_level['geom'] = from_shape(
            unary_union([ shape(f['geometry']) for f in features ]),
            srid=4326
        )

        # Start values with the properties of each GeoJSON Feature.
        # There overwrite as they iterate. Finally they are overridden
        # with the passed in "values".
        values = {}
        for f in features:
            values.update(f['properties'])
        if 'values' in value:
            values.update(value['values'])

        for k, v in value.items():
            if k not in tops:
                # All HSTORE values need to be strings or None
                v = v if v is not None else None
                values[k] = make_valid_string(v)
                del top_level[k]  # Remove from the top level

        if 'reftime' not in top_level:
            top_level['reftime'] = top_level['time']

        # All HSTORE values need to be strings
        values = {
            k: make_valid_string(x) if x is not None else None
            for k, x in values.items()
        }

        top_level['time'] = top_time.isoformat()
        top_level['reftime'] = dtparse(top_level['reftime']).replace(tzinfo=pytz.utc).isoformat()
        top_level['values'] = values
        top_level['payload'] = payload

        # Throw away non-column data
        top_level = self.match_columns(top_level)
        # Remove None to use the defaults defined in the table definition
        return key, { k: v for k, v in top_level.items() if v is not None }


class GenericFloat(BaseMap):

    @property
    def schema(self):
        return [
            sql.Column('id',       sql.Integer, sql.Sequence(self.sequence_name), primary_key=True),
            sql.Column('uid',      sql.String, index=True),
            sql.Column('gid',      sql.String, default='', index=True),
            sql.Column('time',     sql.DateTime(timezone=True), index=True),
            sql.Column('reftime',  sql.DateTime(timezone=True), index=True),
            sql.Column('lat',      DOUBLE_PRECISION, index=True),
            sql.Column('lon',      DOUBLE_PRECISION, index=True),
            sql.Column('z',        DOUBLE_PRECISION, index=True),
            sql.Column('geom',     Geometry('POINT', srid=4326)),
            sql.Column('values',   HSTORE, default={}),
            sql.Column('payload',  JSONB, default={}),
            sql.Index(
                self.unique_index_name,
                'uid',
                'gid',
                'time',
                'lat',
                'lon',
                'z',
                unique=True,
            ),
            sql.UniqueConstraint(
                'uid',
                'gid',
                'time',
                'lat',
                'lon',
                'z',
                name=self.upsert_constraint_name
            )
        ]

    def message_to_values(self, key, value):
        payload = payload_parse(value)

        top_time = dtparse(value['time']).replace(tzinfo=pytz.utc)
        apply_start_end_filter(
            top_time,
            self.filters.get('start_date'),
            self.filters.get('end_date')
        )

        value['lat'] = float(value['lat'])
        value['lon'] = float(value['lon'])
        pt = Point(value['lon'], value['lat'])
        value['geom'] = from_shape(pt, srid=4326)

        if not value['values']:
            value['values'] = {}
        value['values']['location_quality'] = get_point_location_quality(pt)
        # All HSTORE values need to be strings
        value['values'] = { k: make_valid_string(x) for k, x in value['values'].items() }

        value['time'] = top_time.isoformat()
        if 'reftime' in value:
            value['reftime'] = dtparse(value['reftime']).replace(tzinfo=pytz.utc).isoformat()
        else:
            value['reftime'] = value['time']

        value['payload'] = payload

        # Throw away non-column data
        value = self.match_columns(value)
        # Remove None to use the defaults defined in the table definition
        return key, { k: v for k, v in value.items() if v is not None }


class AreteData(GenericFloat):

    def message_to_values(self, key, value):

        values_copy = value.copy()

        # Remove some randoms
        removes = ['not_decoded', 'Compressed_Data']
        for r in removes:
            if r in value['json']:
                del values_copy['json'][r]

        payload = payload_parse(values_copy)
        values = flatten(values_copy)

        # Time - use float timestamp and fall back to Iridium
        reftime = datetime.fromtimestamp(values['headers_iridium_ts'], pytz.utc)
        # TODO: There is no status_ts yet, but this is here for
        # if one does show up eventually
        if values.get('headers_status_ts'):
            timestamp = datetime.fromtimestamp(values['headers_status_ts'], pytz.utc)
        else:
            timestamp = reftime

        apply_start_end_filter(
            timestamp,
            self.filters.get('start_date'),
            self.filters.get('end_date')
        )

        # Location - Use values locations and fall back to Iridium
        inprecise_location = True
        latdeg = float(values['headers_location_latitude_degrees'])
        latmin = float(values['headers_location_latitude_minutes'])
        latdd = latdeg + (latmin / 60)

        londeg = float(values['headers_location_longitude_degrees'])
        lonmin = float(values['headers_location_longitude_minutes'])
        londd = londeg + (lonmin / 60)

        if 'json_Full_ll' in values and isinstance(values['json_Full_ll'], list):
            # Older style positions in early data
            latdd = values['json_Full_ll'][0]
            londd = values['json_Full_ll'][1]
            inprecise_location = False
        elif 'json_position_latitude' in values and 'json_position_longitude' in values:
            # Newer style posititions as of Dec 11, 2019
            latdd = values['json_position_latitude']
            londd = values['json_position_longitude']
            inprecise_location = False

        top_level = {
            'uid':     str(values['headers_imei']),
            'gid':     None,
            'time':    timestamp.isoformat(),
            'reftime': reftime.isoformat(),
            'lat':     latdd,
            'lon':     londd,
            'z':       None,
            'payload': payload
        }
        pt = Point(top_level['lon'], top_level['lat'])
        top_level['geom'] = from_shape(pt, srid=4326)

        # Set additional values
        values['location_quality'] = get_point_location_quality(pt, inprecise_location=inprecise_location)
        values['mfr'] = 'arete'

        # All HSTORE values need to be strings
        values = {
            k: make_valid_string(x) if x is not None else None
            for k, x in values.items()
        }

        fullvalues = {
            **top_level,
            'values': {
                **values
            }
        }

        # Throw away non-column data
        fullvalues = self.match_columns(fullvalues)
        # Remove None to use the defaults defined in the table definition
        return key, { k: v for k, v in fullvalues.items() if v is not None }


class NumurusData(GenericFloat):

    def message_to_values(self, key, value):
        payload = payload_parse(value)

        values = flatten(value)

        top_time = dtparse(values['timestamp']).replace(tzinfo=pytz.utc)
        apply_start_end_filter(
            top_time,
            self.filters.get('start_date'),
            self.filters.get('end_date')
        )

        top_level = {
            'uid':     values['imei'],
            'gid':     None,
            'time':    top_time.isoformat(),
            'reftime': dtparse(values['navsat_fix_time']).replace(tzinfo=pytz.utc).isoformat(),
            'lat':     values['latitude'],
            'lon':     values['longitude'],
            'z':       None,
            'payload': payload,
        }
        pt = Point(top_level['lon'], top_level['lat'])
        top_level['geom'] = from_shape(pt, srid=4326)

        skips = [
            # No easy way to represent this as a flat dict. We can write a db view to extract this
            # data from the `payload` if required.
            'data_segment_data_product_pipeline',
            'data_segment_data_segment_data_product_pipeline'
        ]

        # Set additional values
        # Lat=91 and Lon=181 should be treated as bad location data
        values['location_quality'] = get_point_location_quality(
            pt,
            disallow_lon=[181],
            disallow_lat=[91]
        )
        values['mfr'] = 'numurus'

        # All HSTORE values need to be strings
        values = {
            k: make_valid_string(x) if x is not None else None
            for k, x in values.items()
            if k not in skips
        }

        fullvalues = {
            **top_level,
            'values': {
                **values
            }
        }

        # Throw away non-column data
        fullvalues = self.match_columns(fullvalues)
        # Remove None to use the defaults defined in the table definition
        return key, { k: v for k, v in fullvalues.items() if v is not None }


class NumurusStatus(GenericFloat):

    def message_to_values(self, key, value):
        payload = payload_parse(value)

        values = flatten(value)

        top_time = dtparse(values['timestamp']).replace(tzinfo=pytz.utc)
        apply_start_end_filter(
            top_time,
            self.filters.get('start_date'),
            self.filters.get('end_date')
        )

        top_level = {
            'uid':     values['imei'],
            'gid':     None,
            'time':    top_time.isoformat(),
            'reftime': dtparse(values['navsat_fix_time']).replace(tzinfo=pytz.utc).isoformat(),
            'lat':     values['latitude'],
            'lon':     values['longitude'],
            'z':       None,
            'payload': payload
        }
        pt = Point(top_level['lon'], top_level['lat'])
        top_level['geom'] = from_shape(pt, srid=4326)

        # Set additional values
        # Lat=91 and Lon=181 should be treated as bad location data
        values['location_quality'] = get_point_location_quality(
            pt,
            disallow_lon=[181],
            disallow_lat=[91]
        )
        values['mfr'] = 'numurus'

        # All HSTORE values need to be strings
        values = {
            k: make_valid_string(x) if x is not None else None
            for k, x in values.items()
        }

        fullvalues = {
            **top_level,
            'values': {
                **values
            }
        }

        # Throw away non-column data
        fullvalues = self.match_columns(fullvalues)
        # Remove None to use the defaults defined in the table definition
        return key, { k: v for k, v in fullvalues.items() if v is not None }


class NwicFloatReports(GenericFloat):

    def message_to_values(self, key, value):
        payload = payload_parse(value)

        values = flatten(value)

        # Time - use float timestamp and fall back to Iridium
        reftime = datetime.fromtimestamp(values['headers_iridium_ts'], pytz.utc)
        timestamp = reftime

        # Try to extract a better timestamp
        for k in ['values_status_ts', 'values_environmental_ts', 'values_mission_ts']:
            if values.get(k):
                timestamp = datetime.fromtimestamp(values[k], pytz.utc)
                break

        apply_start_end_filter(
            timestamp,
            self.filters.get('start_date'),
            self.filters.get('end_date')
        )

        # Location - Use values locations and fall back to Iridium
        inprecise_location = True
        latdeg = float(values['headers_location_latitude_degrees'])
        latmin = float(values['headers_location_latitude_minutes'])
        latdd = latdeg + (latmin / 60)

        londeg = float(values['headers_location_longitude_degrees'])
        lonmin = float(values['headers_location_longitude_minutes'])
        londd = londeg + (lonmin / 60)

        if values.get('values_longitude') and values.get('values_latitude'):
            latdd = values['values_latitude']
            londd = values['values_longitude']
            inprecise_location = False

        top_level = {
            'uid':     str(values['headers_imei']),
            'gid':     None,
            'time':    timestamp.isoformat(),
            'reftime': reftime.isoformat(),
            'lat':     latdd,
            'lon':     londd,
            'z':       None,
            'payload': payload
        }
        pt = Point(top_level['lon'], top_level['lat'])
        top_level['geom'] = from_shape(pt, srid=4326)

        # Set additional values
        values['location_quality'] = get_point_location_quality(pt, inprecise_location=inprecise_location)

        # All HSTORE values need to be strings
        values = {
            k: make_valid_string(x) if x is not None else None
            for k, x in values.items()
        }

        fullvalues = {
            **top_level,
            'values': {
                **values
            }
        }

        # Throw away non-column data
        fullvalues = self.match_columns(fullvalues)
        # Remove None to use the defaults defined in the table definition
        return key, { k: v for k, v in fullvalues.items() if v is not None }
